"use strict";
/**
 * Copyright (c) Microsoft Corporation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.RecorderApp = void 0;
const os = require("os");
const fs = require("fs");
const path = require("path");
const util = require("util");
const progress_1 = require("../../progress");
const playwright_1 = require("../../playwright");
const events_1 = require("events");
const chromium_1 = require("../../chromium/chromium");
const readFileAsync = util.promisify(fs.readFile);
class RecorderApp extends events_1.EventEmitter {
    constructor(page) {
        super();
        this.setMaxListeners(0);
        this._page = page;
    }
    async close() {
        await this._page.context().close();
    }
    async _init() {
        const icon = await readFileAsync(require.resolve('../../../web/recorder/app_icon.png'));
        const crPopup = this._page._delegate;
        await crPopup._mainFrameSession._client.send('Browser.setDockTile', {
            image: icon.toString('base64')
        });
        await this._page._setServerRequestInterceptor(async (route) => {
            if (route.request().url().startsWith('https://playwright/')) {
                const uri = route.request().url().substring('https://playwright/'.length);
                const file = require.resolve('../../../web/recorder/' + uri);
                const buffer = await readFileAsync(file);
                await route.fulfill({
                    status: 200,
                    headers: [
                        { name: 'Content-Type', value: extensionToMime[path.extname(file)] }
                    ],
                    body: buffer.toString('base64'),
                    isBase64: true
                });
                return;
            }
            await route.continue();
        });
        await this._page.exposeBinding('dispatch', false, (_, data) => this.emit('event', data));
        this._page.once('close', () => {
            this.emit('close');
            this._page.context().close().catch(e => console.error(e));
        });
        await this._page.mainFrame().goto(new progress_1.ProgressController(), 'https://playwright/index.html');
    }
    static async open() {
        const recorderPlaywright = playwright_1.createPlaywright(true);
        const context = await recorderPlaywright.chromium.launchPersistentContext('', {
            ignoreAllDefaultArgs: true,
            args: [
                ...chromium_1.DEFAULT_ARGS,
                `--user-data-dir=${path.join(os.homedir(), '.playwright-app')}`,
                '--remote-debugging-pipe',
                '--app=data:text/html,',
                `--window-size=300,800`,
            ],
            noDefaultViewport: true
        });
        const controller = new progress_1.ProgressController();
        await controller.run(async (progress) => {
            await context._browser._defaultContext._loadDefaultContextAsIs(progress);
        });
        const [page] = context.pages();
        const result = new RecorderApp(page);
        await result._init();
        return result;
    }
    async setMode(mode) {
        await this._page.mainFrame()._evaluateExpression(((mode) => {
            window.playwrightSetMode(mode);
        }).toString(), true, mode, 'main').catch(() => { });
    }
    async setPaused(paused) {
        await this._page.mainFrame()._evaluateExpression(((paused) => {
            window.playwrightSetPaused(paused);
        }).toString(), true, paused, 'main').catch(() => { });
    }
    async setSource(text, language) {
        await this._page.mainFrame()._evaluateExpression(((param) => {
            window.playwrightSetSource(param);
        }).toString(), true, { text, language }, 'main').catch(() => { });
    }
    async bringToFront() {
        await this._page.bringToFront();
    }
}
exports.RecorderApp = RecorderApp;
const extensionToMime = {
    '.css': 'text/css',
    '.html': 'text/html',
    '.jpeg': 'image/jpeg',
    '.js': 'application/javascript',
    '.png': 'image/png',
    '.ttf': 'font/ttf',
    '.svg': 'image/svg+xml',
    '.webp': 'image/webp',
    '.woff': 'font/woff',
    '.woff2': 'font/woff2',
};
//# sourceMappingURL=recorderApp.js.map